#include "domlette.h"

/*Internal Interfaces*/


/*External Interfaces*/

static char PyDOMImplementation_HasFeature__doc__[] =
"See if this implementation supports a given feature";
static PyObject *PyDOMImplementation_HasFeature(PyObject *self, PyObject *args)
{
/*     _features = {'CORE' : 2.0, */
/*                  'XML' : 2.0, */
/*                  } */

/*     def hasFeature(self, feature, version=''): */
/*         fv = self._features.get(string.upper(feature)) */
/*         if fv: */
/*             if version: */
/*                 try: */
/*                     return float(version) == fv */
/*                 except: */
/*                     return 0 */
/*             return 1 */
/*         return 0 */


  char *feature;
  char *version;
  PyObject *res;

  if(!PyArg_ParseTuple(args,"ss:hasFeature",&feature,&version))
    return NULL;

  if (strcasecmp(feature,"core") ) {
    res = PyInt_FromLong(0);
  }
  else if (atoi(version) <= 2.0 ) {
    res = PyInt_FromLong(1);
  } else {
    res = PyInt_FromLong(0);
  }
  return res;
}


static char PyDOMImplementation_CreateDocument__doc__[] =
"Create a new document instance";
static PyObject *PyDOMImplementation_CreateDocument(PyObject *self,
						    PyObject *args)
{
  PyObject *arg0, *arg1, *doctype, *documentURI = NULL;
  PyObject *namespaceURI, *qualifiedName;
  PyDocumentObject *doc;
  unsigned long index;

  if(!PyArg_ParseTuple(args,"OOO|O:createDocument",
		       &arg0, &arg1, &doctype, &documentURI))
    return NULL;

  /* validate the arguments */
  if (arg0 != Py_None) {
    namespaceURI = PyUnicode_FromObject(arg0);
    if (!namespaceURI) {
      if (PyErr_ExceptionMatches(PyExc_TypeError))
        PyErr_Format(PyExc_TypeError,
                     "namespaceURI must be None, unicode or string, %.80s found.",
                     arg0->ob_type->tp_name);
      return NULL;
    }
  } else {
    namespaceURI = arg0;
    Py_INCREF(Py_None);
  }

  if (arg1 != Py_None) {
    qualifiedName = PyUnicode_FromObject(arg1);
    if (!qualifiedName) {
      if (PyErr_ExceptionMatches(PyExc_TypeError))
        PyErr_Format(PyExc_TypeError,
                     "qualifiedName must be None, unicode or string, %.80s found.",
                     arg1->ob_type->tp_name);
      Py_DECREF(namespaceURI);
      return NULL;
    }
  } else {
    qualifiedName = arg1;
    Py_INCREF(Py_None);
  }

  if (doctype != Py_None) {
    PyErr_SetString(g_notSupportedErr, "doctype must be None for Domlettes");
    Py_DECREF(namespaceURI);
    Py_DECREF(qualifiedName);
    return NULL;
  }

  if (documentURI) {
    documentURI = PyUnicode_FromObject(documentURI);
    if (!documentURI) {
      if (PyErr_ExceptionMatches(PyExc_TypeError))
        PyErr_Format(PyExc_TypeError,
                     "documentURI must unicode or string, %.80s found.",
                     documentURI->ob_type->tp_name);
      Py_DECREF(namespaceURI);
      Py_DECREF(qualifiedName);
      return NULL;
    }
  }

  doc = Document_New(&index, documentURI);

  /* See if we need to add a documentElement */
  if (qualifiedName != Py_None) {
    PyObject *prefix, *localName;
    PyElementObject *elem;

    SplitQName(qualifiedName, &prefix, &localName);
    elem = Document_CreateElementNS(doc, namespaceURI, prefix, localName, &index);

    Py_DECREF(prefix);
    Py_DECREF(localName);

    /* This will add references as needed so we must remove the new
     * reference from the CreateElementNS call.
     */
    Node_AppendChild((PyNodeObject *)doc, (PyNodeObject *)elem);
    Py_DECREF(elem);
  }

  Py_DECREF(namespaceURI);
  Py_DECREF(qualifiedName);
  Py_XDECREF(documentURI);

  return (PyObject *)doc;
}


static char PyDOMImplementation_CreateRootNode__doc__[] =
"Create a new document instance";

static PyObject *PyDOMImplementation_CreateRootNode(PyObject *self,
						    PyObject *args)
{
  PyObject *documentURI = NULL;
  unsigned long index;

  if(!PyArg_ParseTuple(args,"|O:createRootNode", &documentURI))
    return NULL;

  if (documentURI) {
    documentURI = PyUnicode_FromObject(documentURI);
    if (!documentURI) {
      if (PyErr_ExceptionMatches(PyExc_TypeError))
        PyErr_Format(PyExc_TypeError,
                     "documentURI must unicode or string, %.80s found.",
                     documentURI->ob_type->tp_name);
      return NULL;
    }
  }

  return (PyObject *)Document_New(&index, documentURI);
}


static struct PyMethodDef DOMImplementation_methods[] = {
  { "hasFeature", PyDOMImplementation_HasFeature, METH_VARARGS,
    PyDOMImplementation_HasFeature__doc__ },
  { "createDocument", PyDOMImplementation_CreateDocument, METH_VARARGS,
    PyDOMImplementation_CreateDocument__doc__ },
  { "createRootNode", PyDOMImplementation_CreateRootNode, METH_VARARGS,
    PyDOMImplementation_CreateRootNode__doc__ },
  {NULL, NULL}      /* sentinel */
};


/*
  Type Interfaces
*/

static PyObject *domimp_getattr(PyDOMImplementationObject *self, char *name)
{
  return Py_FindMethod(DOMImplementation_methods, (PyObject *)self, name);
}

static int domimp_setattr(PyDOMImplementationObject *self, char *name, PyObject *v)
{
  /* Set attribute 'name' to value 'v'. v==NULL means delete */
  if (v == NULL) {
    PyErr_Format(PyExc_AttributeError,
                 "Cannot delete attribute '%.400s' on '%.50s' object",
                 name, self->ob_type->tp_name);
  } else {
    PyErr_Format(PyExc_AttributeError,
                 "Cannot set attribute '%.400s' on '%.50s' object",
                 name, self->ob_type->tp_name);
  }
  return -1;
}

static void
domimp_dealloc(PyDOMImplementationObject *node)
{
  PyMem_DEL(node);
}


static PyObject *domimp_repr(PyDOMImplementationObject *domimp)
{
  char buf[256];
  sprintf(buf, "<cDOMImplementation at %p>", domimp);
  return PyString_FromString(buf);
}


PyTypeObject PyDomletteDOMImplementation_Type = {
    PyObject_HEAD_INIT(0)
    0,
    "cDomlette.DOMImplementation",
    sizeof(PyDOMImplementationObject),
    0,
    (destructor)domimp_dealloc,   /*tp_dealloc*/
    (printfunc)0,  /*tp_print*/
    (getattrfunc)domimp_getattr,  /*tp_getattr*/
    (setattrfunc)domimp_setattr,  /*tp_setattr*/
    0,                          /*tp_compare*/
    (reprfunc)domimp_repr,          /*tp_repr*/
    0,                          /*tp_as_number*/
    0,              /*tp_as_sequence*/
    0,              /*tp_as_mapping*/
    0,                             /*tp_hash*/
    0,          /*tp_call*/
    0,          /*tp_str*/
    0,                      /*tp_getattro*/
    0,          /*tp_setattro*/
};

