
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include "domlette.h"
#include "reader.h"
#include "getallns.h"

/* needed for API verification */
#include "expat.h"

PyObject *g_xmlNamespace = NULL;
PyObject *g_xmlnsNamespace = NULL;

PyObject *g_emptyAttributes = NULL;
PyObject *g_emptyChildNodes = NULL;
PyObject *g_implementation = NULL;

/* Exceptions */
PyObject *g_errorObject = NULL;
PyObject *g_indexSizeErr = NULL;
PyObject *g_domstringSizeErr = NULL;
PyObject *g_hierarchyRequestErr = NULL;
PyObject *g_wrongDocumentErr = NULL;
PyObject *g_invalidCharacterErr = NULL;
PyObject *g_noDataAllowedErr = NULL;
PyObject *g_noModificationAllowedErr = NULL;
PyObject *g_notFoundErr = NULL;
PyObject *g_notSupportedErr = NULL;
PyObject *g_inuseAttributeErr = NULL;
PyObject *g_invalidStateErr = NULL;
PyObject *g_syntaxErr = NULL;
PyObject *g_invalidModificationErr = NULL;
PyObject *g_namespaceErr = NULL;
PyObject *g_invalidAccessErr = NULL;


/*
  These are the external interfaces
*/

#define UNICODE_FROM_CHAR(str) PyUnicode_DecodeASCII((str), strlen(str), NULL);

#define CHECK_COUNT(num,where) if (doc->ob_refcnt != (num)) { \
    sprintf(errorBuf,"%s expected %d ref count, found %d\n",(where),(num),doc->ob_refcnt); \
    PyErr_SetString(PyExc_MemoryError,errorBuf); \
  return NULL; \
}

/* The test tree XML:
test_xml = """<?xml version = "1.0"?>
<?xml-stylesheet href="addr_book1.xsl" type="text/xml"?>
<docelem xmlns:ft="http://fourthought.com">
  <child foo='bar'>Some Text</child>
  <!--A comment-->
  <ft:nschild ft:foo="nsbar">Some More Text</ft:nschild>
  <appendChild/>
</docelem>"""
*/

PyObject *PyTestTree(PyObject * self, PyObject *args)
{
  /*Build a test tree*/
  char errorBuf[256];
  PyDocumentObject *doc;
  PyProcessingInstructionObject *pi;
  PyElementObject *documentElement, *element;
  PyTextObject *text;
  PyCommentObject *comment;
  PyAttrObject *attr;
  PyObject *target, *data, *localName, *namespaceURI, *prefix, *value;

  unsigned long index = 0;

  doc = Document_New(&index, NULL);
  CHECK_COUNT(1,"Doc Creation");

  /* Add processing instruction */
  target = UNICODE_FROM_CHAR("xml-stylesheet");
  data = UNICODE_FROM_CHAR("href=\"addr_book1.xsl\" type=\"text/xml\"");
  pi = Document_CreateProcessingInstruction(doc, target, data, &index);
  Py_DECREF(data);
  Py_DECREF(target);
  Node_AppendChild((PyNodeObject *)doc, (PyNodeObject *)pi);
  /*We are all done with pi, so DECREF it once*/
  Py_DECREF(pi);

  CHECK_COUNT(3, "PI Creation");

  /* Add documentElement */
  namespaceURI = Py_None;
  prefix = Py_None;
  localName = UNICODE_FROM_CHAR("docelem");
  documentElement = Document_CreateElementNS(doc, namespaceURI, prefix, localName, &index);
  Py_DECREF(localName);
  Node_AppendChild((PyNodeObject *)doc, (PyNodeObject *)documentElement);

  CHECK_COUNT(5,"Doc Elem Creation");

  /* Add documentElement namespace declartion */
  namespaceURI = UNICODE_FROM_CHAR(XMLNS_NAMESPACE);
  localName = UNICODE_FROM_CHAR("ft");
  prefix = UNICODE_FROM_CHAR("xmlns");
  value = UNICODE_FROM_CHAR("http://fourthought.com");
  attr = Element_SetAttributeNS(documentElement, namespaceURI, localName, 
                                prefix, value);
  Py_DECREF(attr);
  Py_DECREF(value);
  Py_DECREF(prefix);
  Py_DECREF(localName);
  Py_DECREF(namespaceURI);

  CHECK_COUNT(6,"NS DECL");

  /* Add text 1 to documentElement */
  data = UNICODE_FROM_CHAR("\n  ");
  text = Document_CreateTextNode(doc, data, &index);
  Py_DECREF(data);
  Node_AppendChild((PyNodeObject *)documentElement, (PyNodeObject *)text);
  /*We are all done with text, so DECREF it once*/
  Py_DECREF(text);
  CHECK_COUNT(7,"1st Text");

  /* Add element 1 to documentElement */
  namespaceURI = Py_None;
  prefix = Py_None;
  localName = UNICODE_FROM_CHAR("child");
  element = Document_CreateElementNS(doc, namespaceURI, prefix, localName, &index);
  Py_DECREF(localName);
  Node_AppendChild((PyNodeObject *)documentElement, (PyNodeObject *)element);
  CHECK_COUNT(8,"First Child");

  /* Add element 1 attribute */
  namespaceURI = Py_None;
  localName = UNICODE_FROM_CHAR("foo");
  prefix = Py_None;
  value = UNICODE_FROM_CHAR("bar");
  attr = Element_SetAttributeNS(element, namespaceURI, localName, prefix, 
                                value);
  Py_DECREF(attr);
  Py_DECREF(localName);
  Py_DECREF(value);
  CHECK_COUNT(9,"First Child Attr");

  /* Add element 1 text */
  data = UNICODE_FROM_CHAR("Some Text");
  text = Document_CreateTextNode(doc, data, &index);
  Py_DECREF(data);
  Node_AppendChild((PyNodeObject *)element, (PyNodeObject *)text);
  /*We are all done with text, so DECREF it once*/
  Py_DECREF(text);
  CHECK_COUNT(10,"First Child Text");

  /*We are all done with element 1, so DECREF it once*/
  Py_DECREF(element);

  /* Add text 2 to documentElement */
  data = UNICODE_FROM_CHAR("\n  ");
  text = Document_CreateTextNode(doc, data, &index);
  Py_DECREF(data);
  Node_AppendChild((PyNodeObject *)documentElement, (PyNodeObject *)text);
  /*We are all done with text, so DECREF it once*/
  Py_DECREF(text);
  CHECK_COUNT(11,"2nd Text");

  /* Add comment to documentElement */
  data = UNICODE_FROM_CHAR("A comment");
  comment = Document_CreateComment(doc, data, &index);
  Py_DECREF(data);
  Node_AppendChild((PyNodeObject *)documentElement, (PyNodeObject *)comment);
  /*We are all done with comment so DECREF it once*/
  Py_DECREF(comment);
  CHECK_COUNT(12,"Comment");

  /* Add text 3 to documentElement */
  data = UNICODE_FROM_CHAR("\n  ");
  text = Document_CreateTextNode(doc, data, &index);
  Py_DECREF(data);
  Node_AppendChild((PyNodeObject *)documentElement, (PyNodeObject *)text);
  /*We are all done with text, so DECREF it once*/
  Py_DECREF(text);
  CHECK_COUNT(13,"3rd Text");

  /* Add element 2 to documentElement */
  namespaceURI = UNICODE_FROM_CHAR("http://fourthought.com");
  prefix = UNICODE_FROM_CHAR("ft");
  localName = UNICODE_FROM_CHAR("nschild");
  element = Document_CreateElementNS(doc, namespaceURI, prefix, localName, &index);
  Py_DECREF(localName);
  Py_DECREF(prefix);
  Py_DECREF(namespaceURI);
  Node_AppendChild((PyNodeObject *)documentElement, (PyNodeObject *)element);
  CHECK_COUNT(14,"2nd Child");

  /* Add element 2 attribute */
  namespaceURI = UNICODE_FROM_CHAR("http://fourthought.com");
  localName = UNICODE_FROM_CHAR("foo");
  prefix = UNICODE_FROM_CHAR("ft");
  value = UNICODE_FROM_CHAR("nsbar");
  attr = Element_SetAttributeNS(element, namespaceURI, localName, prefix, 
                                value);
  Py_DECREF(attr);
  Py_DECREF(value);
  Py_DECREF(prefix);
  Py_DECREF(localName);
  Py_DECREF(namespaceURI);
  CHECK_COUNT(15,"2nd Child Attr");

  /* Add element 2 text node */
  data = UNICODE_FROM_CHAR("Some More Text");
  text = Document_CreateTextNode(doc, data, &index);
  Py_DECREF(data);
  Node_AppendChild((PyNodeObject *)element, (PyNodeObject *)text);
  /*We are all done with text, so DECREF it once*/
  Py_DECREF(text);
  CHECK_COUNT(16,"2nd Child Text");

  /*We are all done with element 2, so DECREF it once*/
  Py_DECREF(element);

  /* Add text 4 to documentElement */
  data = UNICODE_FROM_CHAR("\n  ");
  text = Document_CreateTextNode(doc, data, &index);
  Py_DECREF(data);
  Node_AppendChild((PyNodeObject *)documentElement, (PyNodeObject *)text);
  /*We are all done with text, so DECREF it once*/
  Py_DECREF(text);
  CHECK_COUNT(17,"4th Text");

  /* Add element 3 to documentElement */
  namespaceURI = Py_None;
  prefix = Py_None;
  localName = UNICODE_FROM_CHAR("appendChild");
  element = Document_CreateElementNS(doc, namespaceURI, prefix, localName, &index);
  Py_DECREF(localName);
  Node_AppendChild((PyNodeObject *)documentElement, (PyNodeObject *)element);
  CHECK_COUNT(18,"Append Child");

  /* We are all done with element 3, so DECREF it once*/
  Py_DECREF(element);

  /* Add text 5 to documentElement */
  data = UNICODE_FROM_CHAR("\n  ");
  text = Document_CreateTextNode(doc, data, &index);
  Py_DECREF(data);
  Node_AppendChild((PyNodeObject *)documentElement, (PyNodeObject *)text);
  /*We are all done with text, so DECREF it once*/
  Py_DECREF(text);
  CHECK_COUNT(19,"5th Text");

  /* We are all done with documentElement, so DECREF it once */
  Py_DECREF(documentElement);

  return (PyObject *)doc;

}

static PyObject *PyStartNodeCounting(PyObject * self, PyObject *args)
{
  g_nodeCount = 0;
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject *PyGetNodeCount(PyObject * self, PyObject *args)
{
  return PyInt_FromLong(g_nodeCount);
}

static PyObject *PyTestRefCounts(PyObject * self, PyObject *args)
{
  PyObject *tester;
  PyDocumentObject *document;
  long temp = 0;

  if(!PyArg_ParseTuple(args,"OO:TestRefCounts", &tester, &document)) 
    return NULL;

  if (PyDocument_Check(document)) {
    Py_INCREF(document);
  }
  else if ((PyObject *)document == Py_None) {
    document = Document_New(&temp, NULL);
  }
  else {
    PyErr_SetString(PyExc_TypeError, "argument 2 must be a cDocument or None");
    return NULL;
  }

  if (!document_test_ref_counts(tester, document, document->ob_refcnt)) {
    Py_DECREF(document);
    return NULL;
  }

  Py_DECREF(document);
  Py_INCREF(Py_None);
  return Py_None;
}

/*The external interface definitions */
static PyMethodDef cDomlettecMethods[] = {

  { "parse", PyParse, METH_VARARGS },
  { "TestTree", PyTestTree,METH_VARARGS},
  { "TestRefCounts", PyTestRefCounts,METH_VARARGS},
  { "StartNodeCounting", PyStartNodeCounting,METH_VARARGS},
  { "GetNodeCount", PyGetNodeCount,METH_VARARGS},
  { "GetAllNs", PyGetAllNs,METH_VARARGS,GetAllNs__doc__},
  /*{ "xupdate", PyXUpdate, METH_VARARGS},*/
  { NULL, NULL }
};

DL_EXPORT(void) initcDomlettec(void) {
  PyObject *module, *dict, *modules, *string;
  XML_Expat_Version version;

  /* ensure that we're going to be using the proper Expat functions */
  version = XML_ExpatVersionInfo();
  if (version.major != XML_MAJOR_VERSION || 
      version.minor != XML_MINOR_VERSION ||
      version.micro != XML_MICRO_VERSION) {
    /* version mismatch */
    PyErr_Format(PyExc_ImportError,
                 "Expat version mismatch; expected %d.%d.%d, found %d.%d.%d",
                 XML_MAJOR_VERSION, XML_MINOR_VERSION, XML_MICRO_VERSION,
                 version.major, version.minor, version.micro);
    return;
  }

    module = Py_InitModule("cDomlettec", cDomlettecMethods);
    if (!module) return;
    dict = PyModule_GetDict(module);

    /* Initialize the types */
    PyDomletteDOMImplementation_Type.ob_type = &PyType_Type;
    PyDomletteDocument_Type.ob_type = &PyType_Type;
    PyDomletteDocumentFragment_Type.ob_type = &PyType_Type;
    PyDomletteElement_Type.ob_type = &PyType_Type;
    PyDomletteText_Type.ob_type = &PyType_Type;
    PyDomletteComment_Type.ob_type = &PyType_Type;
    PyDomletteProcessingInstruction_Type.ob_type = &PyType_Type;
    PyDomletteAttr_Type.ob_type = &PyType_Type;

    /* add our implementation */
    g_implementation = (PyObject *)PyObject_New(PyDOMImplementationObject,
                                                &PyDomletteDOMImplementation_Type);
    if (PyDict_SetItemString(dict, "implementation", g_implementation) != 0)
      return;

    g_errorObject = PyErr_NewException("cDomlettec.error", NULL, NULL);
    if (PyDict_SetItemString(dict, "error", g_errorObject) != 0)
      return;

    if (PyDict_SetItemString(dict, "DOMImplementationType", 
                             (PyObject *)&PyDomletteDOMImplementation_Type) != 0)
      return;
    if (PyDict_SetItemString(dict, "DocumentType",
                             (PyObject *)&PyDomletteDocument_Type) != 0)
      return;
    if (PyDict_SetItemString(dict, "DocumentFragmentDocumentType", 
                             (PyObject *)&PyDomletteDocumentFragment_Type) != 0)
      return;
    if (PyDict_SetItemString(dict, "ElementType", 
                             (PyObject *)&PyDomletteElement_Type) != 0)
      return;
    if (PyDict_SetItemString(dict, "TextType", 
                             (PyObject *)&PyDomletteText_Type) != 0)
      return;
    if (PyDict_SetItemString(dict, "CommentType",
                             (PyObject *)&PyDomletteComment_Type) != 0)
      return;
    if (PyDict_SetItemString(dict, "ProcessingInstructionType", 
                             (PyObject *)&PyDomletteProcessingInstruction_Type) != 0)
      return;
    if (PyDict_SetItemString(dict, "AttrType", 
                             (PyObject *)&PyDomletteAttr_Type) != 0)
      return;

    /* PyInt_FromLong returns a new reference */
    g_elementNodeType = PyInt_FromLong(ELEMENT_NODE);
    if (!g_elementNodeType) return;
    g_attributeNodeType = PyInt_FromLong(ATTRIBUTE_NODE);
    if (!g_attributeNodeType) return;
    g_textNodeType = PyInt_FromLong(TEXT_NODE);
    if (!g_textNodeType) return;
    g_cdataSectionNodeType = PyInt_FromLong(CDATA_SECTION_NODE);
    if (!g_cdataSectionNodeType) return;
    g_entityReferenceNodeType = PyInt_FromLong(ENTITY_REFERENCE_NODE);
    if (!g_entityReferenceNodeType) return;
    g_entityNodeType = PyInt_FromLong(ENTITY_NODE);
    if (!g_entityNodeType) return;
    g_processingInstructionNodeType = PyInt_FromLong(PROCESSING_INSTRUCTION_NODE);
    if (!g_processingInstructionNodeType) return;
    g_commentNodeType = PyInt_FromLong(COMMENT_NODE);
    if (!g_commentNodeType) return;
    g_documentNodeType = PyInt_FromLong(DOCUMENT_NODE);
    if (!g_documentNodeType) return;
    g_documentTypeNodeType = PyInt_FromLong(DOCUMENT_TYPE_NODE);
    if (!g_documentTypeNodeType) return;
    g_documentFragmentNodeType = PyInt_FromLong(DOCUMENT_FRAGMENT_NODE);
    if (!g_documentFragmentNodeType) return;
    g_notationNodeType = PyInt_FromLong(NOTATION_NODE);
    if (!g_notationNodeType) return;

    /* PyUnicode_DecodeASCII returns a new reference */
    g_textNodeName = PyUnicode_DecodeASCII("#text", 5, NULL);
    if (!g_textNodeName) return;
    g_cdataSectionNodeName = PyUnicode_DecodeASCII("#cdata-section", 14, NULL);
    if (!g_cdataSectionNodeName) return;
    g_commentNodeName = PyUnicode_DecodeASCII("#comment", 8, NULL);
    if (!g_commentNodeName) return;
    g_documentNodeName = PyUnicode_DecodeASCII("#document", 9, NULL);
    if (!g_documentNodeName) return;
    g_documentFragmentNodeName = PyUnicode_DecodeASCII("#document-fragment", 18, NULL);
    if (!g_documentFragmentNodeName) return;

    Py_INCREF(Py_None);
    g_emptyAttributes = Py_None;
    g_emptyChildNodes = PyList_New(0);
    if (!g_emptyChildNodes) return;

    string = PyString_FromString(XML_NAMESPACE);
    if (!string) return;
    g_xmlNamespace = PyUnicode_FromObject(string);
    Py_DECREF(string);
    if (!g_xmlNamespace) return;
      
    string = PyString_FromString(XMLNS_NAMESPACE);
    if (!string) return;
    g_xmlnsNamespace = PyUnicode_FromObject(string);
    Py_DECREF(string);
    if (!g_xmlnsNamespace) return;
      
    /* Import our exceptions */
    modules = PyImport_GetModuleDict(); /* borrowed reference */
    module = PyDict_GetItemString(modules, "xml.dom"); /* borrowed reference */
    Py_XINCREF(module);
    if (!module) {
      module = PyImport_ImportModule("xml.dom"); /* new reference */
      if (!module) return;
    }

    /* PyObject_GetAttrString returns a new reference */
    g_indexSizeErr = PyObject_GetAttrString(module,"IndexSizeErr");
    if (!g_indexSizeErr) return;

    if (PyObject_HasAttrString(module, "DomstringSizeErr")) {
      /* Python 2.1+ and PyXML 0.8+ */
      g_domstringSizeErr = PyObject_GetAttrString(module, "DomstringSizeErr");
    } else {
      /* PyXML < 0.8 */
      g_domstringSizeErr = PyObject_GetAttrString(module, "DOMStringSizeErr");
    }
    if (!g_domstringSizeErr) return;

    g_hierarchyRequestErr = PyObject_GetAttrString(module,"HierarchyRequestErr");
    if (!g_hierarchyRequestErr) return;

    g_wrongDocumentErr = PyObject_GetAttrString(module,"WrongDocumentErr");
    if (!g_wrongDocumentErr) return;

    g_invalidCharacterErr = PyObject_GetAttrString(module,"InvalidCharacterErr");
    if (!g_invalidCharacterErr) return;

    g_noDataAllowedErr = PyObject_GetAttrString(module,"NoDataAllowedErr");
    if (!g_noDataAllowedErr) return;

    g_noModificationAllowedErr = PyObject_GetAttrString(module,"NoModificationAllowedErr");
    if (!g_noModificationAllowedErr) return;

    g_notFoundErr = PyObject_GetAttrString(module,"NotFoundErr");
    if (!g_notFoundErr) return;

    g_notSupportedErr = PyObject_GetAttrString(module,"NotSupportedErr");
    if (!g_notSupportedErr) return;

    g_inuseAttributeErr = PyObject_GetAttrString(module,"InuseAttributeErr");
    if (!g_inuseAttributeErr) return;

    g_invalidStateErr = PyObject_GetAttrString(module,"InvalidStateErr");
    if (!g_invalidStateErr) return;

    g_syntaxErr = PyObject_GetAttrString(module,"SyntaxErr");
    if (!g_syntaxErr) return;

    g_invalidModificationErr = PyObject_GetAttrString(module,"InvalidModificationErr");
    if (!g_invalidModificationErr) return;

    g_namespaceErr = PyObject_GetAttrString(module,"NamespaceErr");
    if (!g_namespaceErr) return;

    g_invalidAccessErr = PyObject_GetAttrString(module,"InvalidAccessErr");
    if (!g_invalidAccessErr) return;

    Py_DECREF(module);
 
    return;
}
