#ifndef DOMLETTE_H
#define DOMLETTE_H

#define NULL_PREFIX_CHECK(s, where) if (PyUnicode_Check(s) && !PyUnicode_GET_SIZE(s)) {\
  PyErr_SetString(g_namespaceErr, where ": Use None instead of '' for no prefix"); \
  return NULL;\
}

#define NULL_NAMESPACE_CHECK(s, where) if (PyUnicode_Check(s) && !PyUnicode_GET_SIZE(s)) {\
  PyErr_SetString(g_namespaceErr, where ": Use None instead of '' for no namespace"); \
  return NULL;\
}

#define XML_NAMESPACE "http://www.w3.org/XML/1998/namespace"
#define XMLNS_NAMESPACE "http://www.w3.org/2000/xmlns/"
#define XINCLUDE_NAMESPACE "http://www.w3.org/2001/XInclude"

#ifdef __cplusplus
extern "C" {
#endif

#if defined(_WIN32) || defined(__WIN32__)
#  define strcasecmp stricmp
#endif

#include "Python.h"
#include "compat.h"
#include "debug.h"
#include "util.h"

#include "node.h"
#include "document.h"
#include "element.h"
#include "attr.h"
#include "documentfragment.h"
#include "domimplementation.h"
#include "text.h"
#include "processinginstruction.h"

/* namespace constants */
extern PyObject *g_xmlNamespace;
extern PyObject *g_xmlnsNamespace;

/* exception objects */
extern PyObject *g_indexSizeErr;
extern PyObject *g_domstringSizeErr;
extern PyObject *g_hierarchyRequestErr;
extern PyObject *g_wrongDocumentErr;
extern PyObject *g_invalidCharacterErr;
extern PyObject *g_noDataAllowedErr;
extern PyObject *g_noModificationAllowedErr;
extern PyObject *g_notFoundErr;
extern PyObject *g_notSupportedErr;
extern PyObject *g_inuseAttributeErr;
extern PyObject *g_invalidStateErr;
extern PyObject *g_syntaxErr;
extern PyObject *g_invalidModificationErr;
extern PyObject *g_namespaceErr;
extern PyObject *g_invalidAccessErr;

#define PyNode_Check(op) (PyDocument_Check(op) || PyElement_Check(op) || PyText_Check(op) || PyProcessingInstruction_Check(op) || PyComment_Check(op) || PyAttr_Check(op) || PyDocumentFragment_Check(op))

#define PyNode_Children_Check(op) (PyDocument_Check(op) || PyElement_Check(op) || PyDocumentFragment_Check(op))

#define PyNode_Childless_Check(op) (PyText_Check(op) || PyProcessingInstruction_Check(op) || PyComment_Check(op))

#define PyDomlette_Check(node) if (!PyNode_Check(node)) { \
  PyErr_SetString(g_wrongDocumentErr, \
		  "Node created from unsupported document implementation"); \
  return NULL; \
}

PyObject *PyTestTree(PyObject * self, PyObject *args);


#ifdef __cplusplus
}
#endif
#endif /* !DOMLETTEOBJECT_H */

