
#include "util.h"
#include "node.h"
#include "domlette.h"
#include "attr.h"

PyObject *GetAllNsUnknown(PyObject *node);
PyObject *GetAllNscDomlette(PyNodeObject *node);

char GetAllNs__doc__[] = \
"Get all of the namespaces defined in scope of this element";

PyObject *PyGetAllNs(PyObject *self, PyObject *args)
{
  PyObject *node;

  if (!PyArg_ParseTuple(args, "O:GetAllNs", &node))
    return NULL;

  if (PyNode_Check(node)) {
    return GetAllNscDomlette((PyNodeObject *)node);
  }
  return GetAllNsUnknown(node);
}


PyObject *GetAllNsUnknown(PyObject *node) {
  PyObject *py_nodeType = PyObject_GetAttrString(node,"nodeType");
  PyObject *nss,*prefix,*namespace,*localName;

  long nodeType;
  int i=0;
  int ctr;
  PyObject *results, *attributes, *attrs, *attrsFast, *xmlns;
  PyObject *attr,*parentNode,*parentNss, *key,*value;

  if (!py_nodeType) return NULL;

  nodeType = PyInt_AsLong(py_nodeType);

  /*Probably should not hard code node types!!!*/
  if (nodeType == 2) {
    attr = PyObject_GetAttrString(node,"ownerElement");
    if (!attr) return NULL;
    results = GetAllNsUnknown(attr);
    Py_DECREF(attr);
    return results;
  }

  nss = PyDict_New();
  prefix = PyUnicode_DecodeASCII("xml",3,NULL);
  PyDict_SetItem(nss,prefix,g_xmlNamespace);
  Py_DECREF(prefix);

  if (nodeType == 1) {
    attributes = PyObject_GetAttrString(node,"attributes");
    if (!attributes) {
      Py_DECREF(nss);
      return NULL;
    }
    attrs = PyObject_CallMethod(attributes,"values","");
    if (!attrs) {
      Py_DECREF(nss);
      Py_DECREF(attributes);
      return NULL;
    }
    attrsFast = PySequence_Fast(attrs,"Invalid Results from keys");
    Py_DECREF(attrs);
    if (!attrsFast) {
      Py_DECREF(nss);
      Py_DECREF(attributes);
      return NULL;
    }
    for (ctr = 0;ctr < PySequence_Size(attrsFast);ctr ++) {
      attr = PySequence_Fast_GET_ITEM(attrsFast,ctr);
      namespace = PyObject_GetAttrString(attr,"namespaceURI");
      localName = PyObject_GetAttrString(attr,"localName");
      prefix = PyObject_GetAttrString(attr,"prefix");
      value = PyObject_GetAttrString(attr,"value");
      xmlns = PyUnicode_DecodeASCII("xmlns",5,NULL);
      if (!namespace || !prefix || !localName || !xmlns || !value) {
	Py_DECREF(nss);
	Py_DECREF(attributes);
	Py_DECREF(attrsFast);
	Py_XDECREF(namespace);
	Py_XDECREF(localName);
	Py_XDECREF(prefix);
	Py_XDECREF(xmlns);
	Py_XDECREF(value);
	return NULL;
      }

      if (!PyObject_Compare(namespace,g_xmlnsNamespace)) {
	/*A namespace declaration */
	if (!PyObject_Compare(localName,xmlns)) {
	  /*Prefix is probably just none but...*/
	  PyDict_SetItem(nss,prefix,value);
	} else if (prefix != Py_None) {
	  /*Prefix is xmlns, localName is the valid xmlns:foo*/
	  PyDict_SetItem(nss,localName,value);
	} else {
	  /*Not sure how it would get here!!!*/
	  PyDict_SetItem(nss,Py_None,value);
	}
      } else {
	/*See if we need to define its NS*/
      }
      Py_DECREF(namespace);
      Py_DECREF(localName);
      Py_DECREF(prefix);
      Py_DECREF(xmlns);

    }

    /*We do this last cause it has priority*/
    namespace = PyObject_GetAttrString(node,"namespaceURI");
    if (!namespace){
      Py_DECREF(nss);
      return NULL;
    }
    if (namespace != Py_None) {
      prefix = PyObject_GetAttrString(node,"prefix");
      if (!prefix) {
	Py_DECREF(namespace);
	Py_DECREF(nss);
	return NULL;
      }
      PyDict_SetItem(nss,prefix,namespace);
      Py_DECREF(prefix);
    }
    Py_DECREF(namespace);

  }

  parentNode = PyObject_GetAttrString(node,"parentNode");
  if (!parentNode) {
    Py_DECREF(nss);
    return NULL;
  }
  if (parentNode != Py_None) {
    parentNss = GetAllNsUnknown(parentNode);
    if (!parentNss) {
      Py_DECREF(parentNode);
      Py_DECREF(nss);
      return NULL;
    }
    i=0;
    while (PyDict_Next(nss, &i, &key, &value)) {
      PyDict_SetItem(parentNss,key,value);
    }
    Py_DECREF(nss);
    nss = parentNss;
  }
  Py_DECREF(parentNode);

  return nss;
}

PyObject *GetAllNscDomlette(PyNodeObject *node) {
  PyObject *nss,*prefix,*xmlns;

  int i=0;
  PyObject *attr_name_tuple;
  PyObject *attro,*key,*value,*parentNss;
  PyAttrObject *attr;
  PyElementObject *element;

  if (PyAttr_Check(node) && node->parentNode != Py_None) {
    return GetAllNsUnknown(node->parentNode);
  }

  nss = PyDict_New();
  prefix = PyUnicode_DecodeASCII("xml",3,NULL);
  PyDict_SetItem(nss,prefix,g_xmlNamespace);
  Py_DECREF(prefix);

  if (PyElement_Check(node)) {
    element = (PyElementObject *)node;

    xmlns = PyUnicode_DecodeASCII("xmlns",5,NULL);
    if (!xmlns) {
      Py_DECREF(nss);
      return NULL;
    }

    /*Walk over the attributes*/
    while (PyDict_Next(element->attributes, &i, &attr_name_tuple, &attro)) {
      attr = (PyAttrObject *)attro;
      if (!PyObject_Compare(attr->namespaceURI,g_xmlnsNamespace)) {
	/*A namespace decl*/
	if (!PyObject_Compare(attr->localName,xmlns)) {
	  PyDict_SetItem(nss,attr->prefix,attr->nodeValue);
	} else if (attr->prefix != Py_None) {
	  PyDict_SetItem(nss,attr->localName,attr->nodeValue);
	} else {
	  PyDict_SetItem(nss,Py_None,attr->nodeValue);
	}
      }
    }

    /*This is done last so it has the highest priority*/
    if (element->namespaceURI != Py_None) {
      PyDict_SetItem(nss,element->prefix,element->namespaceURI);
    }

  }

  if (node->parentNode != Py_None) {
    parentNss = GetAllNscDomlette((PyNodeObject *)node->parentNode);
    if (!parentNss) {
      Py_DECREF(nss);
      return NULL;
    }
    i=0;
    while (PyDict_Next(nss, &i, &key, &value)) {
      PyDict_SetItem(parentNss,key,value);
    }
    Py_DECREF(nss);
    nss=parentNss;
  }
  
  return nss;
}



  /*    ELEMENT_NODE                = 1
    ATTRIBUTE_NODE              = 2
    TEXT_NODE                   = 3
    CDATA_SECTION_NODE          = 4
    ENTITY_REFERENCE_NODE       = 5
    ENTITY_NODE                 = 6
    PROCESSING_INSTRUCTION_NODE = 7
    COMMENT_NODE                = 8
    DOCUMENT_NODE               = 9
    DOCUMENT_TYPE_NODE          = 10
    DOCUMENT_FRAGMENT_NODE      = 11
    NOTATION_NODE               = 12*/

/*    nss = {'xml': XML_NAMESPACE}
    if node.nodeType == Node.ATTRIBUTE_NODE and node.ownerElement:
        return PyGetAllNs(node.ownerElement)
    if node.nodeType == Node.ELEMENT_NODE:
        if node.namespaceURI:
            nss[node.prefix] = node.namespaceURI
        for attr in node.attributes.values():
            # xmlns="uri" or xmlns:foo="uri"
            if attr.namespaceURI == XMLNS_NAMESPACE:
                if attr.localName == 'xmlns':
                    #Not swapped around
                    nss[attr.prefix] = attr.value
                # xmlns:foo="uri"
                elif attr.prefix:
                    # "foo" prefix -> uri
                    nss[attr.localName] = attr.value
                # xmlns="uri"
                else:
                    # default ns -> uri
                    nss[EMPTY_PREFIX] = attr.value
            # A prefixed attribute name may have had its namespace URI set
            # by the DOM without there having been an xmlns:prefix="uri"
            # that would have been detected by the code above
            elif attr.namespaceURI:
                nss[attr.prefix] = attr.namespaceURI
    if node.parentNode:
        #Inner NS/Prefix mappings take precedence over outer ones
        parent_nss = PyGetAllNs(node.parentNode)
        parent_nss.update(nss)
        nss = parent_nss
	return nss*/







