#ifndef __STATE_MACHINE__
#define __STATE_MACHINE__

#include "expat.h"
#include "ds_primitives.h"
#include "ds_tools.h"

#ifdef __cplusplus
extern "C" {
#endif

/* States */

#define NULL_STATE 0

/* Not really needed, I think */
#define START_STATE 1

#define PARSE_STREAM_STATE 2

#define ENTITY_REF_CALLBACK 10
#define START_ELEMENT_CALLBACK 11
#define END_ELEMENT_CALLBACK 12
#define CHARACTER_DATA_CALLBACK 13
#define COMMENT_CALLBACK 14
#define PI_CALLBACK 15
#define START_NS_SCOPE_CALLBACK 16
#define END_NS_SCOPE_CALLBACK 17

#define ERROR_STATE 30000
#define FIRST_USER_STATE 100

/* Events */

#define PARSE_RESUME_EVENT 1
#define ENTITY_REF_EVENT 10
#define START_ELEMENT_EVENT 11
#define END_ELEMENT_EVENT 12
#define CHARACTER_DATA_EVENT 13
#define COMMENT_EVENT 14
#define PI_EVENT 15
#define START_NS_SCOPE_EVENT 16
#define END_NS_SCOPE_EVENT 17

#define XPTR_MATCH_EVENT 50
#define XPTR_CLOSE_EVENT 52

#define ERROR_EVENT 30000


#define MAX_STATE 65536

typedef struct StateTableEntry_ {
  /* the transitions consists of alternating event and next state in pairs
     until 0 indicates the end */
  short *transitions;
  void (*handler)(void *state, void *params);
  void *params;
} StateTableEntry;


typedef struct State_ {
  short current;
  /* make sure you use calloc to init this to 0 */
  StateTableEntry *table[MAX_STATE];
  char *errorMessage;
} State;


  void transit(void *userState, short event);
  void plainTransitHandler(void *userState, void *params);
  short stateLookupTransition(void *userState, short event);
  int addTransition(void *userState, short from, short event, short next);
  StateTableEntry *newStateTableEntry(void *userState, short id);
  void resetState(void *userState, short new);
  int initializeStateTable(void *userState);
  int destroyStateTable(void *userState);

#define signalError(state, message) { \
  state->core.errorMessage = strdup(message); \
  transit(state, ERROR_EVENT); \
}

#define signalPythonError(state) transit(state, ERROR_EVENT);

#ifdef __cplusplus
}
#endif

#endif

