#include "Python.h"
#include "util.h"

int TranslateNames(PyObject **namespaceUri, PyObject **qualifiedName,
		   PyObject **prefix, PyObject **localName)
{
  if (*namespaceUri == Py_None) {
    Py_INCREF(Py_None);
  } 
  else {
    *namespaceUri = PyUnicode_FromObject(*namespaceUri);
    if (*namespaceUri == NULL) 
      return 0;
  }

  *qualifiedName = PyUnicode_FromObject(*qualifiedName);
  if (*qualifiedName == NULL) {
    Py_DECREF(*namespaceUri);
    return 0;
  }

  SplitQName(*qualifiedName, prefix, localName);
  if (*prefix == NULL || *localName == NULL) {
    Py_DECREF(*namespaceUri);
    Py_DECREF(*qualifiedName);
    Py_XDECREF(*prefix);
    Py_XDECREF(*localName);
    return 0;
  }

  /* Success! */
  return 1;
}

void AddInternCtr(PyObject *obj, PyObject *dict)
{
  PyObject *value;

  if (obj != Py_None) {
    value = PyDict_GetItem(dict, obj);
    if (value == NULL) {
      /*Add it  also adds a ref*/
      value = PyInt_FromLong(1);
      PyDict_SetItem(dict, obj, value);
      Py_DECREF(value);
    } else {
      value = PyInt_FromLong(PyInt_AsLong(value) + 1);
      PyDict_SetItem(dict, obj, value);
      Py_DECREF(value);
    }
  }
}

int TestRefCount(PyObject *tester,PyObject *obj,int expected, const char* name)
{
  char buf[500];
  PyObject *retval = PyObject_Repr(obj);
  if (retval == NULL) return 0;
  sprintf(buf, "%.200s equal to %.200s", name, PyString_AS_STRING(retval));
  Py_DECREF(retval);

  retval = PyObject_CallMethod(tester, "startTest", "s", buf);
  if (retval == NULL) return 0;
  Py_DECREF(retval);

  retval = PyObject_CallMethod(tester, "compare", "ll", (long)expected, obj->ob_refcnt);
  if (retval == NULL) return 0;
  Py_DECREF(retval);

  retval = PyObject_CallMethod(tester, "testDone", NULL);
  if (retval == NULL) return 0;
  Py_DECREF(retval);
  return 1;
}

