#include "parse_event_handler.h"
#include "expat_module.h"
#include "domlette.h"
#include "xptr.h"


#define SIGNAL_STATE_MACHINE_ERROR { \
    signalError(state, "Internal failure"); \
    return 0; \
  }


unsigned long handle_xpointer(ParserState *state, PyObject *xptr_spec)
{
  int i, length, new_start_state=0;
  /*unsigned long xi_start_state=PARSE_STREAM_STATE;*/
  PyObject *states;

    /* Set up base state (equivalent of PARSE_STREAM_STATE for this XInclude) */
    states = PyObject_GetAttrString(xptr_spec, "states");
    length = PySequence_Length(states);
    for (i=0; i < length; i++){
      PyObject *tuple = PySequence_GetItem(states, i);
      int prev_base = PyInt_AsLong(PySequence_GetItem(tuple, 0));
      int prev_start = prev_base == PARSE_STREAM_STATE ? PARSE_STREAM_STATE : prev_base + 1;
      int prev_end = prev_start + 1;
      int base = PyInt_AsLong(PySequence_GetItem(tuple, 1));
      int start = PyInt_AsLong(PySequence_GetItem(tuple, 2));
      int end = PyInt_AsLong(PySequence_GetItem(tuple, 3));
      PyObject *params = PySequence_GetItem(tuple, 4);
      StateTableEntry *ste;

      if (!new_start_state) new_start_state = base;
      if ((ste = newStateTableEntry(state, base)) == NULL)
        SIGNAL_STATE_MACHINE_ERROR;
      /* Set up start element handlers */
      if ((ste = newStateTableEntry(state, start)) == NULL)
        SIGNAL_STATE_MACHINE_ERROR;
      ste->handler = xptrStartElemHandler;
      ste->params = params;
      if (i==length-1){
        /* Final state */
        if (!addTransition(state, start, XPTR_MATCH_EVENT, START_ELEMENT_CALLBACK))
          SIGNAL_STATE_MACHINE_ERROR;
      }
      if (!addTransition(state, prev_start, XPTR_MATCH_EVENT, base))
        SIGNAL_STATE_MACHINE_ERROR;
      if (!addTransition(state, base, START_NS_SCOPE_EVENT, base))
        SIGNAL_STATE_MACHINE_ERROR;
      if (!addTransition(state, base, COMMENT_EVENT, base))
        SIGNAL_STATE_MACHINE_ERROR;
      if (!addTransition(state, base, PI_EVENT, base))
        SIGNAL_STATE_MACHINE_ERROR;
      if (!addTransition(state, base, CHARACTER_DATA_EVENT, base))
        SIGNAL_STATE_MACHINE_ERROR;
      if (!addTransition(state, base, START_ELEMENT_EVENT, start))
        SIGNAL_STATE_MACHINE_ERROR;
      if (!addTransition(state, base, PARSE_RESUME_EVENT, base))
        SIGNAL_STATE_MACHINE_ERROR;
      if (!addTransition(state, start, PARSE_RESUME_EVENT, base))
        SIGNAL_STATE_MACHINE_ERROR;

      /* Set up end element handlers */
      if ((ste = newStateTableEntry(state, end)) == NULL)
        SIGNAL_STATE_MACHINE_ERROR;
      /*
        ste->handler = endXptrStartElemHandler;
      */
      ste->handler = plainTransitHandler;
      if ((ste->params = (int *)calloc(1, sizeof(int))) == NULL) {
        /* Exception: Out of memory */
        return 0;
      }
      *(int *)ste->params = i ? base : prev_base;
      if (!addTransition(state, base, END_ELEMENT_EVENT, base))
        SIGNAL_STATE_MACHINE_ERROR;
      if (!addTransition(state, base, XPTR_CLOSE_EVENT, prev_end))
        SIGNAL_STATE_MACHINE_ERROR;
      if (i==length-1){
        /* Final state */
        if (!addTransition(state, END_ELEMENT_CALLBACK, XPTR_CLOSE_EVENT, base))
          SIGNAL_STATE_MACHINE_ERROR;
      }
    }
    return new_start_state;
}


void xptrStartElemHandler(void *userState, void *params)
{
  ParserState *state = (ParserState *)userState;
  PyObject *criterion, *criterion_code, *criteria = (PyObject *)params;
  PyObject *test_uri, *test_local, *test_value, *value;
  PyObject *target_count, *counter_list, *counter, *one=PyInt_FromLong(1);
  UniversalName *un;
  int i, length, match=1;
  Py_INCREF(one);

  length = PySequence_Length(criteria);
  for (i=0; i < length && match; i++) {
    criterion = PySequence_GetItem(criteria, i);
    criterion_code = PySequence_GetItem(criterion, 0);
    switch (PyInt_AsLong(criterion_code)) {

    case ELEMENT_MATCH:
      test_uri = PySequence_GetItem(criterion, 1);  /* new reference */
      test_local = PySequence_GetItem(criterion, 2); /* new reference */

      /* Call build universal name to get the current NS. */
      un = buildUniversalName(state, state->expat_name);
      if (!un) {
        /*Should have already transitioned to an error*/
        Py_DECREF(test_uri);
        Py_DECREF(test_local);
        Py_DECREF(criterion_code);
        Py_DECREF(criterion);
        return;
      }
      if (PyObject_Compare(un->uri, test_uri) ||
          PyObject_Compare(un->local, test_local)){
        match = 0;
      }

      Py_DECREF(test_uri);
      Py_DECREF(test_local);
      destroyUniversalName(un);
      break;

    case ELEMENT_COUNT:
      target_count = PySequence_GetItem(criterion, 1); /* new reference */
      counter_list = PySequence_GetItem(criterion, 2); /* new reference */
      counter = PySequence_GetItem(counter_list, 0); /* new reference */

      if (PyObject_Compare(counter, target_count)){
        match = 0;
      }
      Py_DECREF(counter);

      counter = PyNumber_InPlaceAdd(counter, one); /* new reference */
      PyList_SetItem(counter_list, 0, counter);

      Py_DECREF(target_count);
      Py_DECREF(counter_list);
      Py_DECREF(counter);
      break;

    case ATTRIBUTE_MATCH:
      test_uri = PySequence_GetItem(criterion, 1);
      test_local = PySequence_GetItem(criterion, 2);
      test_value = PySequence_GetItem(criterion, 3);

      if (test_uri == Py_None) {
        /* bare name; OK to find the attribute directly */
        value = PyDict_GetItem(state->expat_atts, test_local); /* borrowed */
      } else {
        int pos = 0;
        PyObject *nssep = PyUnicode_FromUnicode(NULL, 1);
        PyObject *expat_name;

        /* Create the unicode object separator character for contains */
        if (!nssep) {
          Py_DECREF(test_uri);
          Py_DECREF(test_local);
          Py_DECREF(test_value);
          Py_DECREF(criterion_code);
          Py_DECREF(criterion);
          return;
        }

        PyUnicode_AS_UNICODE(nssep)[0] = EXPAT_NSSEP;
        
        while (PyDict_Next(state->expat_atts, &pos, &expat_name, &value)) {
          /* only test attribute names that have a namespace */
          if (PyUnicode_Contains(expat_name, nssep) > 0) {
            un = buildUniversalName(state, expat_name);
            if (!un) {
              /* error */
              Py_DECREF(nssep);
              Py_DECREF(test_uri);
              Py_DECREF(test_local);
              Py_DECREF(test_value);
              Py_DECREF(criterion_code);
              Py_DECREF(criterion);
              return;
            }
            if (!PyObject_Compare(un->uri, test_uri) &&
                !PyObject_Compare(un->local, test_local)) {
              /* we found the attribute */
              destroyUniversalName(un);
              break;
            }     
            destroyUniversalName(un);
          }
        }

        Py_DECREF(nssep);
        if (pos == PyDict_Size(state->expat_atts)) {
          /* we made it all the way through without a match */
          value = NULL;
        }
      }

      if (!value || PyObject_Compare(value, test_value)) {
        match = 0;
      }
      Py_DECREF(test_uri);
      Py_DECREF(test_local);
      Py_DECREF(test_value);
      break;

    default:
      signalError(state, "Internal error in Xpointer");
    }
    Py_DECREF(criterion_code);
    Py_DECREF(criterion);
  }

  if (match) {
    *state->curr_elem_depth_event = XPTR_CLOSE_EVENT;
    transit(state, XPTR_MATCH_EVENT);
  }
}


void endXptrStartElemHandler(void *userState, void *params)
{
  ParserState *state = (ParserState *)userState;
  PyObject *xptr_params = (PyObject *)params;
  PyObject *code;
  PyObject *test_uri;
  PyObject *test_local;
  UniversalName *un;

  code = PySequence_GetItem(xptr_params, 0);
  switch (PyInt_AsLong(code)) {
  case ELEMENT_MATCH:
    test_uri = PySequence_GetItem(xptr_params, 1);
    test_local = PySequence_GetItem(xptr_params, 2);
    /* Call build universal name to get the current NS. */
    un = buildUniversalName(state, state->expat_name);
    if (!un) {
      /*Should have already transitioned to an error*/
      Py_DECREF(test_local);
      Py_DECREF(test_uri);
      Py_DECREF(code);
      return;
    }
    if (!PyObject_Compare(un->uri, test_uri) &&
        !PyObject_Compare(un->local, test_local)){
      transit(state, XPTR_MATCH_EVENT);
    }
    destroyUniversalName(un);
    Py_DECREF(test_local);
    Py_DECREF(test_uri);
    break;
  default:
    signalError(state, "Internal error in Xpointer");
  }

  Py_DECREF(code);
}


