#include "splitqname.h"

static char SplitQName__doc__[] = \
"SplitQName(qualifiedName) -> (prefix, localName)\n\
\n\
where 'qualifiedName' is a QName according to XML Namespaces 1.0\n\
<http://www.w3.org/TR/REC-xml-names>.\n\
returns the name parts according to the spec.";

PyObject *PySplitQName(PyObject *self, PyObject *args)
{
  PyObject *qualifiedName, *localName, *prefix, *splitName;

  if (!PyArg_ParseTuple(args, "O:SplitQName", &qualifiedName))
    return NULL;

  qualifiedName = PyUnicode_FromObject(qualifiedName);
  if (!qualifiedName) return NULL;

  SPLITQNAME(qualifiedName, &prefix, &localName);
  Py_DECREF(qualifiedName);
  if (!prefix || !localName) {
    Py_XDECREF(prefix);
    Py_XDECREF(localName);
    return NULL;
  }

  splitName = PyTuple_New(2);
  if (!splitName) {
    Py_DECREF(prefix);
    Py_DECREF(localName);
    return NULL;
  }
  /* steals references */
  PyTuple_SET_ITEM(splitName, 0, prefix);
  PyTuple_SET_ITEM(splitName, 1, localName);
  return splitName;
}


/*The external interface definitions */
static PyMethodDef splitQNamecMethods[] = {

  { "SplitQName", PySplitQName, METH_VARARGS, SplitQName__doc__ },
  { NULL, NULL }
};

DL_EXPORT(void) initSplitQNamec(void) {
  PyObject *module;

  module = Py_InitModule("SplitQNamec", splitQNamecMethods);
  if (!module) return;
 
  return;
}



